#!/usr/bin/env python
import os
import re
import json
from tempfile import gettempdir
from math import floor
from telegram import Update
from telegram.ext import Application, CommandHandler, MessageHandler, filters, ContextTypes
import openai

from .openai_singleton import OpenAIClientSingleton
from .initialize import INIT_BOT_CONFIG, init_structure, init_bot_config, init_bot_prompt, init_bot_commands
from .conversation import Conversation
from .tokenGPT import TokenGPT
from .message_handlers import handle_greetings, handle_common_queries, handle_url_ask
from .utils import read_yaml, read_text, exact_word_match, generate_error_path, log_error


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class TelegramBot:
    # Show all the commands available by the bot using the TelegramCommands.txt file
    async def tele_commands(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        reply = ''
        with open(init_bot_commands('commands.txt'), "r", encoding="utf-8") as file:
            for line in file:
                reply += line
        await update.message.reply_text(reply)

    # Start command only runs by the 'bot_owner' specified in configuration file 
    async def tele_start_command(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        uname = update.message.from_user.username
        if uname == self.telegram['owner']:
            greeting_text = f"Oh, hello {update.message.from_user.first_name}! Let me get to work!"
            await update.message.reply_text(greeting_text)
            self.started = True
            self.GPTOnline = True
        else:
            await update.message.reply_text("Sorry, but I'm off the clock at the moment.")

    # Stop command formally closes out the polling loop
    async def tele_stop_command(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        uname = update.message.from_user.username
        if uname == self.telegram['owner']:
            self.GPTOnline = False
            await update.message.reply_text("Sure thing boss, cutting out!")
        else:
            await update.message.reply_text("Sorry, I can't do that for you.")

    # Let bot know to call the user by a different name other than the Telegram user name
    # noinspection RegExpRedundantEscape,PyArgumentList
    async def tele_nick_command(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        # Must follow the format of "/nick <nickname>" using regular expression
        result = re.search(r'^\/nick[ \n]+([ \S]+)', update.message.text)
        if result is not None:
            prompt = f"Please refer to me by my nickname, {result.group(1).strip()}, rather than my user name."
            await update.message.reply_text(await self.tele_handle_response(text=prompt, update=update))
        else:
            await update.message.reply_text("Please provide a valid nickname after the command like \"/nick B0b #2\".")

    # Remove the whole conversation including past session logs for peace of mind
    async def tele_forget_command(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        uname = update.message.from_user.username
        if uname in self.conversations:
            self.conversations[uname].clear_interaction()
            del self.conversations[uname]  # Delete object as well
            await update.message.reply_text("My memories of all our conversations are wiped!")
        else:
            await update.message.reply_text(
                "My apologies, but I don't recall any conversations with you, or you"
                " already asked me to forget about you. Either way, nice to meet you!"
            )

    # Responses
    async def tele_handle_response(self, text: str, update: Update) -> str:
        # Before we handle messages, ensure a user has /started us
        # Starting ensures we get some kind of user account details for logging
        not_started_reply = "I'd love to chat, but please wait as I haven't started up yet!"
        if not self.started:
            return not_started_reply

        # For a new session, track if the user has conversed with the Telegram bot before
        # Username is consistent across restarts and different conversation instances
        uname = update.message.from_user.username
        if uname not in self.conversations:
            self.conversations[uname] = Conversation(
                uname,
                self.telegram['username'],
                self.chatgpt['prompt'],
                self.chatgpt['chat_model']
            )
            # If there are past conversations via logs, load by 50% threshold of tokens
            self.conversations[uname].get_past_interaction(
                floor(self.chatgpt['prune_threshold'] / 2)
            )

        # Add the user's message to our conversation
        self.conversations[uname].add_user_message(text)

        # Check if the user is asking about a [URL]
        url_match = re.search(r'\[http(s)?://\S+]', text)

        # Form the assistant's message based on low level easy stuff or send to GPT
        # OpenAI's Responses API relies on the maximum number of tokens the selected model can support
        reply = not_started_reply
        if handle_greetings(text):
            reply = handle_greetings(text)
        elif handle_common_queries(text):
            reply = handle_common_queries(text)
        elif url_match:
            # URL content is passed into another model to summarize (GPT-4 preferred)
            await update.message.reply_text("Sure, give me a moment to look at that URL...")
            reply = await handle_url_ask(text, self.chatgpt['url_model'])
        elif self.GPTOnline:
            # This is essentially the transition point between quick Telegram replies and GPT
            response = await self.gpt_completion(uname)
            reply = response.choices[0].message.content

        # Calculate the total token count of our conversation messages via tiktoken
        token_count = self.conversations[uname].get_message_token_count()

        # If the user is getting closer to the bot's token threshold, warn the first time
        if token_count > self.chatgpt['prune_back_to'] and uname not in self.token_warning:
            reply += ("\n\n"
                      "By the way, our conversation will soon reach my token limit, so I may"
                      " start forgetting some of our older exchanges. Would you like me to"
                      " summarize our conversation so far to keep the main points alive?"
                      )
            self.token_warning[uname] = True

        # Add assistant's message to the user's conversation
        self.conversations[uname].add_assistant_message(reply)

        # Truncate older messages if the conversation is above the bot's token threshold
        if token_count > self.chatgpt['prune_threshold']:
            self.conversations[uname].prune_conversation(self.chatgpt['prune_back_to'])

        return reply

    # Handles the Telegram side of the message, discerning between Private and Group conversation
    async def tele_handle_message(self, update: Update, context=ContextTypes.DEFAULT_TYPE):
        message_type: str = update.message.chat.type  # PM or Group Chat
        message_text: str = update.message.text
        message_print = f"User {update.message.from_user.username} in {message_type} chat ID {update.message.chat.id}"

        # If it's a group text, only reply if the bot is named
        # The real magic of how the bot behaves is in tele_handle_response()
        if message_type == 'supergroup' or message_type == 'group':
            if exact_word_match(self.telegram['username'], message_text):
                print(message_print)
                new_text: str = message_text.replace(self.telegram['username'], '').strip()
                response: str = await self.tele_handle_response(text=new_text, update=update)
            elif (
                exact_word_match(self.telegram['nickname'], message_text) or
                exact_word_match(self.telegram['initials'], message_text)
            ):
                print(message_print)
                response: str = await self.tele_handle_response(text=message_text, update=update)
            else:
                return
        elif message_type == 'private':
            print(message_print)
            response: str = await self.tele_handle_response(text=message_text, update=update)
        else:
            return
        await update.message.reply_text(response)

    # Handle errors caused on the Telegram side
    async def tele_error(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        if update.message.from_user.username is None:
            # If the user doesn't have a Telegram username, we can't really do anything
            await update.message.reply_text('Add a username to your Telegram account so that I can talk to you!')
        else:
            log_error(context.error, error_type='Telegram', error_filename=self.error_log)
            await update.message.reply_text("Sorry, I ran into an error! Please contact my creator.")

    # Read the GPT Conversation so far
    @staticmethod
    def gpt_read_interactions(file_path: str):
        with open(file_path, 'r') as interaction_log:
            lines = interaction_log.readlines()
        formatted_messages = [json.loads(line) for line in lines]
        return formatted_messages

    # Get the OpenAI response based on bot configuration
    async def gpt_completion(self, uname: str):
        try:
            client = OpenAIClientSingleton.get_instance()
            response = await client.chat.completions.create(
                model=self.chatgpt['chat_model'],
                messages=self.conversations[uname].messages,
            )
            return response

        except openai.AuthenticationError as e:
            # Handle authentication error
            log_error(e, error_type='OpenAI-Authentication', error_filename=self.error_log)
        except openai.BadRequestError as e:
            # Handle invalid request error
            if re.search(r'maximum context.+reduce the length', str(e)):
                # Remove older messages and try again since the model's maximum token limit reached
                print(f"Response to {uname} reached maximum context length, pruning conversation")
                self.conversations[uname].prune_conversation(self.chatgpt['prune_back_to'])
                return await self.gpt_completion(uname)
            else:
                # Another error is actually invalid to investigate
                log_error(e, error_type='OpenAI-BadRequest', error_filename=self.error_log)
        except openai.APIConnectionError as e:
            # Handle API connection error
            log_error(e, error_type='OpenAI-APIConnection', error_filename=self.error_log)
        except Exception as e:
            # Catch any other unexpected exceptions
            log_error(e, error_type='Other', error_filename=self.error_log)

    # The main polling "loop" the user interacts with via Telegram
    def start_polling(self):
        print(f"TeLLMgramBot {self.telegram['username']} polling...")
        self.telegram['app'].run_polling(poll_interval=self.telegram['pollinterval'])
        print(f"TeLLMgramBot {self.telegram['username']} polling ended.")

    # Initialization
    def __init__(self,
                 bot_username   = INIT_BOT_CONFIG['bot_username'],
                 bot_owner      = INIT_BOT_CONFIG['bot_owner'],
                 bot_name       = INIT_BOT_CONFIG['bot_name'],
                 bot_nickname   = INIT_BOT_CONFIG['bot_nickname'],
                 bot_initials   = INIT_BOT_CONFIG['bot_initials'],
                 chat_model     = INIT_BOT_CONFIG['chat_model'],
                 url_model      = INIT_BOT_CONFIG['url_model'],
                 token_limit    = INIT_BOT_CONFIG['token_limit'],
                 persona_temp   = INIT_BOT_CONFIG['persona_temp'],
                 persona_prompt = INIT_BOT_CONFIG['persona_prompt']
                 ):
        # First provide the main structure if not already there
        init_structure()

        # Set up our variables
        self.error_log = generate_error_path()
        self.started = False
        self.GPTOnline = False
        self.token_warning = {}  # Determines whether user has reached token limit by OpenAI model
        self.conversations = {}  # Provides Conversation class per user based on bot response

        # Get Telegram Spun Up
        # noinspection PyDictCreation
        self.telegram = {
            'owner'        : bot_owner,
            'username'     : bot_username,
            'nickname'     : bot_nickname,
            'initials'     : bot_initials,
            'pollinterval' : 3
        }
        self.telegram['app'] = Application.builder().token(os.environ['TELLMGRAMBOT_TELEGRAM_API_KEY']).build()

        # Add our handlers for Commands, Messages, and Errors
        self.telegram['app'].add_handler(CommandHandler('help', self.tele_commands))
        self.telegram['app'].add_handler(CommandHandler('start', self.tele_start_command))
        self.telegram['app'].add_handler(CommandHandler('stop', self.tele_stop_command))
        self.telegram['app'].add_handler(CommandHandler('nick', self.tele_nick_command))
        self.telegram['app'].add_handler(CommandHandler('forget', self.tele_forget_command))
        self.telegram['app'].add_handler(MessageHandler(filters.TEXT, self.tele_handle_message))
        self.telegram['app'].add_error_handler(self.tele_error)

        # Get our LLM spun up with defaults if not defined by user input
        # Tokens as integers measure the length of conversation messages
        self.chatgpt = {
            'name'        : bot_name,
            'prompt'      : persona_prompt,
            'chat_model'  : chat_model,
            'url_model'   : url_model,
            'token_limit' : token_limit or TokenGPT(chat_model).max_tokens(),
            'temperature' : persona_temp or 1.0,
            'top_p'       : 0.9
        }
        # Set a rounded-down integer to prune a lengthy conversation by 500 tokens
        # Note if the upper limit is below 500, the lower limit is set to 0
        self.chatgpt['prune_threshold'] = floor(0.95 * self.chatgpt['token_limit'])
        self.chatgpt['prune_back_to'] = max(0, self.chatgpt['prune_threshold'] - 500)

    # Sets TeLLMgramBot object based on its YAML configuration and prompt files
    # noinspection PyMethodParameters
    def set(config_file='config.yaml', prompt_file='test_personality.prmpt'):
        # First provide the main structure if not already there
        init_structure()

        # Ensure both bot configuration and prompt files are defined and readable
        config = read_yaml(init_bot_config(config_file))
        prompt = read_text(init_bot_prompt(prompt_file))

        # Check any configuration values missing and apply default values:
        for parameter, value in INIT_BOT_CONFIG.items():
            if parameter == 'persona_prompt':
                # Apply initial prompt if not defined
                if not prompt:
                    prompt = value
                    print(f"File '{prompt_file}' is empty, set default prompt '{prompt}'")
            elif parameter not in config:
                # Apply initial configuration parameter with default if not defined
                config[parameter] = value
                if value:
                    print(f"Configuration '{parameter}' not defined, set to '{value}'")

        # Apply parameters to bot:
        return TelegramBot(
            bot_username   = config['bot_username'],
            bot_owner      = config['bot_owner'],
            bot_name       = config['bot_name'],
            bot_nickname   = config['bot_nickname'],
            bot_initials   = config['bot_initials'],
            chat_model     = config['chat_model'],
            url_model      = config['url_model'],
            token_limit    = config['token_limit'],
            persona_temp   = config['persona_temp'],
            persona_prompt = prompt
        )



# Utility Functions
import os
import re
import yaml
from datetime import datetime
from tempfile import gettempdir
from traceback import format_exc

# File Name Friendly Timestamp
def get_timestamp() -> str:
    now = datetime.now()
    timestamp = now.strftime("%Y-%m-%d_%H-%M-%S")
    return timestamp

# File Name Friendly UserName
def get_safe_username(name: str) -> str:
    # Remove leading '@' and any other special characters to exclude
    return re.sub(r'[^\w\s]', '', name.lstrip('@'))

# Basic open text file function
def open_file(file_path: str):
    with open(file_path, 'r', encoding='utf-8') as file:
        return file.read()

# Reads the file per line in reverse order
# Source: https://thispointer.com/python-read-a-file-in-reverse-order-line-by-line/
def read_reverse_order(file_path: str):
    with open(file_path, 'r', encoding='utf-8') as file:
        # Get all lines in a file as list to reverse order
        lines = file.readlines()
        lines = [line.strip() for line in lines]
        lines = reversed(lines)
        return lines

# Read a YAML-formatted file, useful for configurations
def read_yaml(file_path: str):
    with open(file_path, 'r', encoding='utf-8') as file:
        data = yaml.safe_load(file)
    return data

# Read a plain text file, useful for system prompts
def read_text(file_path: str) -> str:
    text = open_file(file_path)
    return text.strip()

# Determine if the given text has the exact word
# Example 'Word': 'Word-cross' ✓ vs. 'word1-cross' X
def exact_word_match(word: str, text: str) -> bool:
    return re.search(rf'\b{word}\b', text) is not None

# Generate a Chat Session Log filename
def generate_filename(bot_name: str, user_name: str) -> str:
    timestamp = get_timestamp()
    return f'{bot_name}-{user_name}-{timestamp}.log'

# Appends text to existing file path
def append_text_file_path(file_path: str, text=""):
    if not text:
        pass
    else:
        try:
            with open(file_path, 'a', encoding='utf-8') as f:
                f.write(f"{text}\n")
        except Exception as e:
            print(f"{type(e).__name__} writing to '{file_path}':\n{format_exc()}")

# Generates a new file for a given directory with optional print name and text
def generate_file_path(directory: str, file: str, text="", print_name="") -> str:
    path = os.path.join(directory, file)
    print_name = "path" if not print_name else print_name 
    if not os.path.exists(path) and not os.path.isfile(path):
        try:
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'w', encoding='utf-8') as f:
                if not text:
                    pass
                else:
                    f.write(text)
            print(f"Created {print_name} file: {path}")
        except Exception as e:
            print(f"{type(e).__name__} generating '{path}':\n{format_exc()}")
    return path

# Generates a location of error log file by env variable vs. temporary location
def generate_error_path(file='tellmgrambot_error.log') -> str:
    env_var = 'TELLMGRAMBOT_ERRORLOGS_PATH'
    if not os.environ.get(env_var):
        os.environ[env_var] = gettempdir()
        print(f"{env_var} initialized with: {os.environ[env_var]}")
    return generate_file_path(os.environ[env_var], file, f"empty {env_var}")

# Log error messages to a file including an error type and timestamp
def log_error(error: Exception or str, error_type: str, error_filename=generate_error_path()):
    text = f"{get_timestamp()} {error_type}: {error}"
    print(text) # Also log into console
    append_text_file_path(error_filename, text)



import os
from openai import AsyncOpenAI


class OpenAIClientSingleton:
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = AsyncOpenAI(api_key=os.environ['TELLMGRAMBOT_OPENAI_API_KEY'])
        return cls._instance



# Provides TeLLMgramBot a way to store and retrieve conversation messages by the OpenAI model
import os
import re
import sys

from .tokenGPT import TokenGPT
from .utils import generate_filename, generate_file_path, generate_error_path
from .utils import get_safe_username, read_reverse_order, append_text_file_path, log_error

# noinspection RegExpRedundantEscape,PyTypeChecker
class Conversation:
    def __init__(self, user_name: str, assist_name: str, system_content: str, system_model="gpt-4o-mini"):
        # Define who's talking for the two-way conversation
        self.user_name = get_safe_username(user_name)
        self.assist_name = assist_name

        # System defines how the assistant will respond to the user by prompt and GPT model
        self.system_content = system_content
        self.system_model = TokenGPT(system_model)
        self.messages = [{"role": "system", "content": system_content}]

        # Define the directory where to store this conversation with log file name
        env_var = 'TELLMGRAMBOT_SESSIONLOGS_PATH'
        if not os.environ.get(env_var):
            sys.exit(f"{env_var} must exist during conversation! Exiting...")
        self.error_log = generate_error_path()
        self.interaction_dir = os.environ[env_var]
        self.interaction_log = generate_filename(self.assist_name, self.user_name)
        if not os.path.exists(os.environ.get(env_var)):
            generate_file_path(self.interaction_dir, self.interaction_log, "", "Conversation")

        # Private boolean determines if conversation includes past messages
        # This will be true if user calls out function get_past_interaction()
        self._has_past_interaction = False

    # Replace the system content as the first message of the conversation
    def set_system_content(self, new_content: str):
        self.system_content = new_content
        self.messages[0] = {"role": "system", "content": new_content}

    # Add user message and write to this conversation's log file
    def add_user_message(self, content: str):
        self.messages.append({"role": "user", "content": content})
        self.write_interaction(role="user", content=content)

    # Add assistant message and write to this conversation's log file
    def add_assistant_message(self, content: str):
        self.messages.append({"role": "assistant", "content": content})
        self.write_interaction(role="assistant", content=content)

    # Clear out current conversation by:
    def clear_messages(self):
        # Any user-assistant messages even from past, except system message
        self.messages = [self.messages[0]]
        self._has_past_interaction = False

        # Log file storing the messages 
        log_path = os.path.join(self.interaction_dir, self.interaction_log)
        if os.path.isfile(log_path):
            with open(log_path, 'w'):
                pass

    # Each message size is determined by the number of tokens via OpenAI's tiktoken library
    # Another way would be to use bytes, but not as accurate since it depends on GPT model
    def get_message_token_count(self):
        return self.system_model.num_tokens_from_messages(self.messages)

    # Write to a new log file the conversation message by role
    # This addresses UTF-8 characters, especially other symbols like integrals
    def write_interaction(self, role: str, content: str):
        log_path = os.path.join(self.interaction_dir, self.interaction_log)
        try:
            text = f'<{role}>: {content}'
            if not os.path.exists(log_path):
                generate_file_path(self.interaction_dir, self.interaction_log, text, "Conversation")
            else:
                append_text_file_path(log_path, text)
        except Exception as e:
            log_error(e, f"{type(e).__name__} on '{log_path}'", self.error_log)

    # Pop messages off the stack until the token size is below threshold
    def prune_conversation(self, token_limit: int):
        self.messages.pop()
        if self.get_message_token_count() > token_limit:
            self.prune_conversation(token_limit)

    # Get all conversation log files of the same user and assistant
    def get_interaction_files(self):
        # Start looping each log file sorted by timestamp
        files = []
        for file in sorted(os.listdir(self.interaction_dir)):
            # Check if full names of user and assistant are in the log name
            if file.startswith(f"{self.assist_name}-{self.user_name}-"):
                files.append(file)
        return files

    # Read logs to store past conversation messages up to a token limit
    def get_past_interaction(self, token_limit: int):
        # First check if the conversation does not have past user-assistant messages
        names = f"User {self.user_name} & Assistant {self.assist_name}"
        if not self._has_past_interaction:
            # Set initial variables if token limit has reached and initial number of messages
            limit_reached = 0
            message_count = len(self.messages)
            token_count = 0

            # Start looping each log file with timestamp in descending order
            for log_file in sorted(self.get_interaction_files(), reverse=True):
                # Skip if the log file is the current conversation log
                if log_file == self.interaction_log:
                    continue

                # Iterate through the log file backwards by line and populate content for each role
                role = ''  # Set role blank, user should start off before assistant
                content = ''  # Set content blank to start off
                for line in read_reverse_order(os.path.join(self.interaction_dir, log_file)):
                    # Remove leading and trailing spaces
                    line = line.strip()

                    # Append content after the stripped line by scenario:
                    # 1.) A new line if there is an empty line with actual content in between.
                    if line == '' and content != '':
                        content = '\n' + content
                    # 2.) The right poriton of the line if the role name by format
                    #     ('<user>: ...', '<assistant>: ...') is defined.
                    result = re.search(r'^\<(user|assistant)\>: (.*)', line)
                    if result is not None:
                        role = result.group(1)
                        content = result.group(2) + content
                    # 3.) The whole line if the role name by format is not defined.
                    else:
                        content = line + content

                    # If the role and content are defined, add both as one message
                    # Otherwise keep adding the content until the role is defined
                    if result is not None:
                        # Since we are going backwards, put it after the first and only system message
                        self.messages.insert(1, {"role": role, "content": content})
                        content = ''  # Reset content for next round

                        # See if the series of messages does actually fit by token limit
                        token_count = self.get_message_token_count()
                        if token_count > token_limit:
                            self.messages.pop(1)  # Remove message inserted recently
                            token_count = self.get_message_token_count()
                            limit_reached = 1
                            break

                # Stop going through logs if the token limit has reached
                if limit_reached:
                    break

            # Print in terminal if there were conversations via session logs
            self._has_past_interaction = True
            if message_count != len(self.messages):
                print(f"{names} had past conversations {'above' if limit_reached else 'within'}"
                      f" the token limit {token_limit}, storing {token_count} token(s)")
            else:
                print(f"{names} are in their first conversation")
        else:
            print(f"{names} had their past conversations defined")

    # Erase every message and log of the user-assistant conversation
    def clear_interaction(self):
        self.clear_messages()
        for file in self.get_interaction_files():
            log_path = os.path.join(self.interaction_dir, file)
            if os.path.isfile(log_path):
                os.remove(log_path)
            else:
                log_error("File not found to delete", f"Error on '{log_path}'", self.error_log)



# Fetches the content of a URL and returns the HTML as a string
# If a specific <div> tag ID is provided, extracts only that content
import os
import re
import asyncio
from typing import Optional
from urllib.parse import urlparse

import httpx
import validators
from bs4 import BeautifulSoup


class InvalidURLException(Exception):
    """Exception raised for invalid URLs."""
    pass


class InsecureURLException(Exception):
    """Exception raised for insecure URLs."""
    pass


class SusURLException(Exception):
    """Exception raised for insecure URLs."""
    pass


class FetchingError(Exception):
    """Exception raised for errors during fetching."""
    pass


async def check_url_safety(url: str, max_retries=5, retry_delay=5) -> Optional[bool]:
    headers = {'x-apikey': os.environ['TELLMGRAMBOT_VIRUSTOTAL_API_KEY']}
    data = {'url': url}

    async with httpx.AsyncClient() as client:
        response = await client.post(url='https://www.virustotal.com/api/v3/urls', headers=headers, data=data)

        if response.status_code == 200:  # OK
            json_response = response.json()
            analysis_id = json_response.get("data", {}).get("id", None)  # Extract the analysis ID from the response
            if analysis_id:
                retries = 0  # Initialize retry counter
                while retries <= max_retries:  # While we haven't exceeded max retries
                    # Now check the analysis results
                    response = await client.get(
                        url=f'https://www.virustotal.com/api/v3/analyses/{analysis_id}',
                        headers=headers
                    )
                    if response.status_code == 200:  # OK
                        json_response = response.json()
                        attributes = json_response.get("data", {}).get("attributes", None)
                        if attributes:
                            status = attributes.get("status", None)
                            if status == "completed":
                                malicious = attributes.get("stats", {}).get("malicious", None)
                                suspicious = attributes.get("stats", {}).get("suspicious", None)
                                if malicious == 0 and suspicious < 3:
                                    return True  # URL is safe
                                else:
                                    print(f"URL is not safe. Malicious: {malicious}, "
                                          f"Suspicious: {suspicious}, Status: {status}")
                                    return False  # URL is not safe
                            elif status == "queued":
                                retries += 1  # Increment retry counter
                                await asyncio.sleep(retry_delay)  # Wait before retrying
                            else:
                                print(f"Unexpected status: {status}")
                                return None
                        else:
                            print("Attributes not found in the response.")
                            return None
                    else:
                        print("Error in fetching analysis results.")
                        return None
            else:
                print("Analysis ID not found in the response.")
                return None
        else:
            print(f"An error occurred: {response.status_code}")
            return None


async def fetch_url(url: str, div_id: Optional[str] = None, div_class: Optional[str] = None) -> Optional[str]:
    try:
        async with httpx.AsyncClient() as client:
            # Validate the URL
            if not validators.url(url):
                print(f"Invalid URL: {url}")
                raise InvalidURLException(f'Invalid URL: {url}')
            elif not urlparse(url).scheme.startswith('https'):
                print(f"URL is not secure: {url}")
                raise InsecureURLException(f'URL is not secure: {url}')
            elif not await check_url_safety(url):
                print(f"URL is not safe: {url}")
                raise SusURLException('URL is not safe or is non-existent!')

            # Fetch the URL if everything is OK
            response = await client.get(url)
            response.raise_for_status()  # Raises an exception for 4xx and 5xx responses

            # Parse the HTML content
            soup = BeautifulSoup(response.text, features='html.parser')

            # If a specific <div> tag ID or class is provided, extract only that content
            specific_div = None
            if div_id:
                specific_div = soup.find(name='div', id=div_id)
            elif div_class:
                specific_div = soup.find(name='div', class_=div_class)

            if specific_div:
                return str(specific_div)  # Returns the HTML of the specified <div> tag
            elif div_id or div_class:
                print(f"Specified <div> tag with ID {div_id} or class {div_class} not found")
                return None  # Returns none but they were expecting to find something
            else:
                return response.text  # Returns the entire HTML content as they didn't specify a <div> tag

    except httpx.HTTPError as err:
        print(f"An unhandled HTTP error occurred while fetching the content from the URL provided: {err}")
        raise FetchingError(
            f'An unhandled HTTP error occurred while fetching the content from  the URL provided: {err}')


# Strips the HTML markup from a string
def strip_html_markup(html_content: str) -> str:
    # Parse the HTML content
    soup = BeautifulSoup(html_content, features='html.parser')

    # Extract the plain text content
    text_content = soup.get_text()

    # Remove excessive blank lines
    cleaned_text = re.sub('\n+', '\n', text_content).strip()

    return cleaned_text



# Defines GPT model parameters and token count of messages using OpenAI's tiktoken library
# Sources:
# > https://platform.openai.com/docs/models
# > https://github.com/openai/openai-python
# > https://platform.openai.com/docs/guides/model-optimization
import re
import tiktoken
from .utils import read_yaml
from .initialize import init_tokenGPT_config

class TokenGPT:
    def __init__(self, openai_model="gpt-4o-mini", yaml_file="tokenGPT.yaml"):
        # The OpenAI model name may also be a fine-tuned model, which has the base model
        # after "ft:" up to the next colon (:), like "ft:gpt-4o-mini:..."
        self.model = re.search("^ft\:([^:]*)", openai_model).group(1) if openai_model.startswith(
            "ft:") else openai_model
        self.config_path = init_tokenGPT_config(yaml_file)

        # Get model parameters by configuration file for the:
        # > Maximum number of tokens
        # > Tokens per message
        # > Tokens per name by model
        self.param = None
        for key, param in read_yaml(self.config_path).items():
            # Set parameters if the configuration key matches either:
            #  > Full OpenAI model name (an exact match to stop searching)
            #  > Base model name after "ft:" up to next colon (:), part of the OpenAI model name
            if key == openai_model or key == self.model:
                self.param = param
                if key == openai_model:
                    break

        # If the parameters are not set, the OpenAI model name is invalid or
        # has an undefined model configuration not set in the YAML file
        if self.param is None:
            raise ValueError(
                f"OpenAI model \"{openai_model}\" is invalid or its base model is not in:\n{self.config_path}")

        # Set token model parameter defaults, unless defined in configuration:
        if 'max_tokens' not in self.param:
            self.param['max_tokens'] = 4097
        if 'tokens_per_message' not in self.param:
            self.param['tokens_per_message'] = 3
        if 'tokens_per_name' not in self.param:
            self.param['tokens_per_name'] = 1

        # Get OpenAI's tiktoken encoding of the base model name
        try:
            self.encoding = tiktoken.encoding_for_model(self.model)
        except Exception:
            fallback = "o200k_base" if re.search("^gpt-[45]", self.model) else "cl100k_base"
            try:
                self.encoding = tiktoken.get_encoding(fallback)
            except Exception:
                self.encoding = tiktoken.get_encoding("cl100k_base")
            print(f"Warning: using {self.encoding.name} encoding for model \"{self.model}\"")

    # Query the maximum amount of tokens possible an OpenAI model can support
    def max_tokens(self) -> int:
        return self.param['max_tokens']

    # Return the number of tokens based on a list of messages
    # Source: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    def num_tokens_from_messages(self, messages: dict[str, str]) -> int:
        num_tokens = 0
        for message in messages:
            num_tokens += self.param['tokens_per_message']
            for key, value in message.items():
                num_tokens += len(self.encoding.encode(value))
                if key == "name":
                    num_tokens += self.param['tokens_per_name']
        num_tokens += 3  # Every reply is primed with <|start|>assistant<|message|>
        return num_tokens



# This script initializes TeLLMgramBot with useful functions
import os
import sys

# Handle relative modules if running this script vs. elsewhere
try:
    from .utils import read_text, generate_file_path
except ImportError:
    from utils import read_text, generate_file_path

INIT_BOT_CONFIG = {
    'bot_username': 'test_bot',
    'bot_owner': '<YOUR USERNAME>',
    'bot_name': 'Test Bot',
    'bot_nickname': 'Testy',
    'bot_initials': 'TB',
    'chat_model': 'gpt-5-mini',
    'url_model': 'gpt-5',
    'token_limit': None,
    'persona_temp': None,
    'persona_prompt': 'You are a test harness bot.'
}

def execution_dir() -> str:
    """
    Initializes the top-level execution's path directory
    """
    execution = os.path.abspath(sys.argv[0])
    if os.path.basename(execution) == 'initialize.py':
        execution = os.path.dirname(execution)
    return os.path.dirname(execution)

# Checks for important directories to configure for TeLLMgramBot
# If each environment variable is not defined, set directory with base execution path:
# - TELLMGRAMBOT_CONFIGS_PATH
# - TELLMGRAMBOT_PROMPTS_PATH
# - TELLMGRAMBOT_ERRORLOGS_PATH
# - TELLMGRAMBOT_SESSIONLOGS_PATH
def init_directories():
    for directory in ['configs', 'prompts', 'errorlogs', 'sessionlogs']:
        env_var = f"TELLMGRAMBOT_{directory.upper()}_PATH"
        if not os.environ.get(env_var):
            os.environ[env_var] = os.path.join(execution_dir(), directory)
            try:
                if not os.path.exists(os.environ[env_var]):
                    os.makedirs(os.environ[env_var])
                    print(f"Created {env_var} directory: {os.environ[env_var]}")
            except Exception as e:
                sys.exit(f"{type(e).__name__} on {env_var} directory '{os.environ[env_var]}': {e}\nExiting...")

# TeLLMgramBot utilizes three API keys by environment variable first then .key file:
# - TELLMGRAMBOT_OPENAI_API_KEY
# - TELLMGRAMBOT_TELEGRAM_API_KEY
# - TELLMGRAMBOT_VIRUSTOTAL_API_KEY
# - TELLMGRAMBOT_KEYS_PATH (otherwise base application path) for key files:
#   + openai.key
#   + telegram.key
#   + virustotal.key 
def init_keys():
    # List each key and URL if it does not exist for more information
    keys = {
        'openai': 'https://platform.openai.com/api-keys',
        'telegram': 'https://core.telegram.org/api',
        'virustotal': 'https://developers.virustotal.com/reference/overview'
    }
    # Ensure appropriate keys can be populated from environment variables or files
    keys_need = 0
    for key, url in keys.items():
        env_var = f"TELLMGRAMBOT_{key.upper()}_API_KEY"
        # If the environment variable is undefined, set by API key file
        if not os.environ.get(env_var):
            path = os.path.join(os.environ.get('TELLMGRAMBOT_KEYS_PATH', execution_dir()), f"{key}.key")
            if not os.path.exists(path):
                # Create a basic *.key file for user to input next run
                generate_file_path(os.path.dirname(path), os.path.basename(path), (
                    f"YOUR {key.upper()} API KEY HERE - {url}\n"
                ), "API key")
                keys_need = 1
                continue
            else:
                # Load each API key by file into its environment variable
                os.environ[env_var] = read_text(path)
                print(f"Loaded {env_var} secret from: {path}")

        # API keys must be defined and not have any spaces
        if not os.environ.get(env_var) or any(c.isspace() for c in os.environ.get(env_var)):
            print(f"{env_var} secret cannot be blank or have any whitelines")
            keys_need = 1

    # Stop program if one of the API keys is needed or has an error
    if keys_need:
        sys.exit("Need appropriate API keys to run TeLLMgramBot! Exiting...")

# Ensure the list of TeLLMgramBot commands is provided for user by a text file
def init_bot_commands(file='commands.txt') -> str:
    env_var = "TELLMGRAMBOT_CONFIGS_PATH"
    try:
        return generate_file_path(os.environ[env_var], file, (
            "TeLLMgramBot commands:\n"
            "/start - Go online (Admin only)\n"
            "/stop - Go offline (Admin only)\n"
            "/nick - Add your nickname to set\n"
            "/forget - Clear your conversations\n"
            "/help - Show this information\n\n"
            "NOTE: Chat history will still show in your Telegram client. You must have a username, which is your "
            "default nickname unless specified like \"/nick B0b #2\".\n"
        ), 'Telegram commands')
    except KeyError:
        sys.exit(f"{env_var} must be defined to create file '{file}'! Exiting...")
    except Exception:
        sys.exit(f"Error while generating {env_var} file '{file}'! Exiting...")

# Ensure bot configuration file is created
def init_bot_config(file='config.yaml') -> str:
    env_var = "TELLMGRAMBOT_CONFIGS_PATH"
    try:
        path = os.path.join(os.environ[env_var], file)
        if not os.path.exists(path):
            # Create a basic TeLLMgramBot configuration by filename
            generate_file_path(os.path.dirname(path), file, "", "bot configuration")
            with open(path, 'w') as f:
                # Add configuration with default values except prompt:
                for parameter, value in INIT_BOT_CONFIG.items():
                    if parameter == 'persona_prompt':
                        next
                    else:  # Write parameter, with optional comment if no value
                        f.write('%s : %s\n' % (
                            parameter.ljust(12),
                            value if value else '# Optional, see README'
                        ))
        return path
    except KeyError:
        sys.exit(f"{env_var} must be defined to create file '{file}'! Exiting...")
    except Exception:
        sys.exit(f"{type(e).__name__} on {env_var} file '{file}': {e}\nExiting...")

# Ensures TokenGPT configuration file is created
def init_tokenGPT_config(file='tokenGPT.yaml') -> str:
    env_var = "TELLMGRAMBOT_CONFIGS_PATH"
    try:
        # Create a basic GPT configuration of token parameters by filename
        return generate_file_path(os.environ[env_var], file, (
            "# Available parameters per base OpenAI model with default values:\n"
            "#   max_tokens: 4097\n"
            "#   tokens_per_message: 3\n"
            "#   tokens_per_name: 1\n"
            "# For OpenAI's updated list of models, please see:\n"
            "#   https://platform.openai.com/docs/models\n"
            "'gpt-4o':\n"
            "  max_tokens: 128000\n"
            "'gpt-4o-mini':\n"
            "  max_tokens: 128000\n"
            "'gpt-5':\n"
            "  max_tokens: 400000\n"
            "'gpt-5-mini':\n" 
            "  max_tokens: 400000\n"
            "'gpt-5-nano':\n"
            "  max_tokens: 400000\n"
        ), "TokenGPT configuration")
    except KeyError:
        sys.exit(f"{env_var} must be defined to create file '{file}'! Exiting...")
    except Exception:
        sys.exit(f"Error while generating {env_var} file '{file}'! Exiting...")

# Ensure prompt file is created that defines bot personality
def init_bot_prompt(file='test_personality.prmpt') -> str:
    env_var = "TELLMGRAMBOT_PROMPTS_PATH"
    try:
        return generate_file_path(os.environ[env_var], file, (
            "You are a test harness bot based on a GPT-5 mini model that can:\n"
            "1. Fetch URLs, provided the user supply a URL in [square brackets].\n"
            "2. Scrub URLs to ensure they are safe for work via the VirusTotal API.\n"
        ), "bot personality prompt")
    except KeyError:
        sys.exit(f"{env_var} must be defined to create file '{file}'! Exiting...")
    except Exception:
        sys.exit(f"Error while generating {env_var} file '{file}'! Exiting...")

# Ensure prompt file is created that defines how the bot will analyze URLS via square brackets []
def init_url_prompt(file='url_analysis.prmpt') -> str:
    env_var = "TELLMGRAMBOT_PROMPTS_PATH"
    try:
        return generate_file_path(os.environ[env_var], file, (
            "The user has provided a URL to perform some level of analysis. You will infer the nature of the "
            "analysis from the user's query.\n\n"
            "The contents of the URL mentioned have already been harvested and cleansed. Note the URL contents "
            "will likely have sections of text that are less relevant to the user's question (headers, footers, "
            "menus, ads, etc.). You will need to ignore those sections of text and focus on the main content of "
            "the page.\n\n"
            "The contents of the URL are shown below:\n"
            "BEGIN URL CONTENTS\n"
            "{url_content}\n"
            "END URL CONTENTS\n"
        ), "URL analysis prompt")
    except KeyError:
        sys.exit(f"{env_var} must be defined to create file '{file}'! Exiting...")
    except Exception:
        sys.exit(f"Error while generating {env_var} file '{file}'! Exiting...")
    return path

# Performs main setup, especially with necessary files
def init_structure():
    init_directories()
    init_keys()
    init_url_prompt()
    init_tokenGPT_config()
    init_bot_commands()

# Run this script if the first time setting up TeLLMgramBot
if __name__ == '__main__':
    init_structure()
    init_bot_config()
    init_bot_prompt()
    print("TeLLMgramBot setup complete!")
